c 
c                         SUBROUTINE : CSCG
c                         -----------------
c
c       (C)omplex (S)ymmetric (C)onjugate (G)radient Algorithm 
c
c       This subroutine attempts a solution of an Ax=b type problem 
c       where A is a known sparse complex symmetric matrix, b is a 
c       known vector and the approximate solution vector is obtained 
c       via a simple complex symmetric conjugate gradient procedure.  
c       The algorithm used here is based on algorithm 10.3-1 in 
c       "Matrix Computations", first edition, by G. Golub and C. 
c       Van Loan, Johns Hopkins Univ. Press, 1983.  The "Lanczos" 
c       tridiagonal matrix in constructed using formula 10.2-14 in 
c       the reference above.
c
c       It is assumed in this routine that the right hand vector is 
c       normalized and the solution vector is initialized at the start
c       of the routine.  The matrix and index arrays are passed to 
c       this routine through the inclusion of the file eprmat.inc.
c
c       Modified by SHL :  Lanczos vector is stored in array 'lnzvec'.
c       They are used in the conversion of the tridiagonal eigenvector
c	into the eigenvector in the original ritz basis.  The role of
c	the old cgltri.f routine (calculation of the Lanczos tridiagonal
c	from the quantities which arise naturally in the course of the
c	CG algorithm) is also incorporated in this routine to resolve
c	the sign ambiguity that comes in when one takes sqrt(complex #).
c	This happens twice here.
c	  i)  converting beta(CG) into beta(LA)
c	  ii) normalizing the residual vector
c	As long as the sign is consistent between these, the resulting
c	eigenvectors and eigenvalues are the same, even though the sign
c	of eigenvectors and beta's of the tridiagonal matrix may be
c 	different from that of LA.
c
c       arguments 
c       ---------
c
c             b     : right hand vector
c             ndim  : number of rows and columns in the matrix 
c             mxcgs : maximum number of conjugate gradient steps allowed
c             cgtol : maximum residual allowed for termination
c             shift : a complex shift value used to avoid extraneous
c                     divisions by zero 
c             ter,r,p,w : work spaces
c      returns
c      -------
c             x     : approximate solution vector 
c             lnzvec: Lanczos vectors that defines the transformation
c                     from the original to the tridiagonal matrix
c             al    : diagonal elements of tridiagonal matrix
c             bl    : off-diagonal elements of tridiagonal matrix
c             ter   : b-A*x
c             ndone : number of cg steps executed (positive if 
c                     converged, negative otherwise)
c             terror: true error of the solution vector estimated
c                     Termination of the CG steps is based on this
c                     true error.
c
c       written by DJS 20-SEP-87
c	modified by SHL 22-JUL-92
c
c       Includes:
c               nlsdim.inc
c               parcom.inc
c               stdio.inc
c               rndoff.inc
c       Uses:
c               zdotu2.f
c               zaypx.f
c               scmvm.f
c               zaxpy2.f
c                               
c**********************************************************************
c
      subroutine cscg(b,ndim,mxcgs,cgtol,shift,
     #   x,lnzvec,al,bl,ter,r,p,w,ndone,terror,trmin,ntrmin)
c
      implicit none
      include 'limits.inc'
      include 'parms.inc'
      include 'stdio.inc'
      include 'rndoff.inc'
c
      integer ndim,mxcgs,ndone,ntrmin
      double precision cgtol,terror,trmin,terror2,error0
      complex*16 alpha,alph1,alph2,beta,x,b,lnzvec,al,bl,r
      dimension x(mxdim),b(mxdim),lnzvec(mxdim,mxstep),
     #          al(mxstep),bl(mxstep),r(mxdim)
c
      integer i,nstpcg
      double precision zero,one,amp,phase,tr,ti,trhalf,tihalf,
     # c,c1,c2,zetabt,zeta,zetaa,w0,w1,w2
      complex*16 tao,taotmp,rho2h,czero,ci,shift,blt,theta,
     # eta,thetab,thetaa,s,s1,s2,zetab,cone
      parameter (zero=0.0D0,one=1.0D0,czero=(0.0D0,0.0D0),
     #           ci=(0.0D0,1.0D0),cone=(1.0D0,0.0D0))
c
      complex*16 w,ter,p,p1,p0,phit,ptmp,ptmp2,stmp,stmp2
      dimension ter(mxdim),p(mxdim),p1(mxdim),p0(mxdim),
     # w(mxdim),phit(mxdim),ptmp(mxdim),ptmp2(mxdim)
c
      complex*16 zdotu2,cal_psnorm
      double precision cal_enorm
      external zdotu2,cal_psnorm,cal_enorm
c
c######################################################################
c
c----------------------------------------------------------------------
c    Initialize r,x and p
c----------------------------------------------------------------------
c
      bl(1)=cal_psnorm(b,ndim)
      do i=1,ndim
        phit(i)= czero
        p(i)= czero
        p0(i)= czero
        p1(i)= czero
        x(i)=czero
        ptmp(i)=b(i)/bl(1)
        ter(i)=b(i) ! for the new auxiliary vector
        r(i)=b(i)    ! ditto
      enddo
      stmp=one
c
      w1=cal_enorm(ptmp,ndim)
      tao=bl(1)*w1
c
c======================================================================
c     begin loop over CG steps
c======================================================================
c
      nstpcg=0
 100  nstpcg=nstpcg+1
c
c----------------------------------------------------------------------
c     calculate beta 
c----------------------------------------------------------------------
c
      if (nstpcg.eq.1) then
        do i=1,ndim
         lnzvec(i,1)=b(i)/bl(1)
         ptmp(i)=lnzvec(i,1)
        enddo
        call scmvm(ptmp,w,ndim)
        call zaxpy2(ptmp,w,shift,ndim)
        al(1)=zdotu2(ptmp,w,ndim)
        do 10 i=1,ndim
 10     phit(i)=w(i)-al(1)*ptmp(i)
      else
        do i=1,ndim
         lnzvec(i,nstpcg)=lnzvec(i,nstpcg)/bl(nstpcg)
         ptmp(i)=lnzvec(i,nstpcg)
        enddo
        call scmvm(ptmp,w,ndim)
        call zaxpy2(ptmp,w,shift,ndim)
        al(nstpcg)=zdotu2(ptmp,w,ndim)
        do i=1,ndim
        phit(i)=w(i)-al(nstpcg)*lnzvec(i,nstpcg)
     #    -bl(nstpcg)*lnzvec(i,nstpcg-1)
        enddo
      end if
c
c----------------------------------------------------------------------
      blt=cal_psnorm(phit,ndim)
c----------------------------------------------------------------------
c     compute variables, theta, eta, zeta, c, s
c----------------------------------------------------------------------
c
      if (nstpcg.eq.1) then
        theta=czero
c        do 40 i=1,ndim
c 40     ptmp(i)=lnzvec(i,1)
c        w1=cal_enorm(ptmp,ndim)
        eta=czero
        zetab=w1*al(1)
      else if (nstpcg.eq.2) then
        theta=czero
        do i=1,ndim
         ptmp(i)=lnzvec(i,nstpcg-1)
         ptmp2(i)=lnzvec(i,nstpcg)
        enddo
        w0=cal_enorm(ptmp,ndim) 
        w1=cal_enorm(ptmp2,ndim)
        eta=c1*w0*bl(2)-dconjg(s1)*w1*al(2)
        zetab=c1*w1*al(2)-s1*w0*bl(2)
      else
        do i=1,ndim
         ptmp(i)=lnzvec(i,nstpcg-1) 
         ptmp2(i)=lnzvec(i,nstpcg)
        enddo
        w0=cal_enorm(ptmp,ndim)
        w1=cal_enorm(ptmp2,ndim)
        theta=dconjg(s2)*w0*bl(nstpcg)
        eta=c1*c2*w0*bl(nstpcg)+dconjg(s1)*w1*al(nstpcg)
        zetab=c1*w1*al(nstpcg)-s1*c2*w0*bl(nstpcg)
      end if
c----------------------------------------------------------------------
c    
c----------------------------------------------------------------------
c
      do i=1,ndim
       ptmp(i)=phit(i)/blt
      enddo
      w2=cal_enorm(ptmp,ndim)
      zetabt=dabs(dreal(zetab)) 
c
      tr=dreal(blt)
      ti=dimag(blt)
      zetaa=sqrt(zetabt+w2*w2*(tr*tr+ti*ti))
c
      if (zetabt.gt.rndoff) then
        zeta=zetaa*zetab/sqrt(zetabt)
      else
        zeta=zetaa
      end if
      c=zetab/zeta
      s=w2*blt/zeta
      stmp=stmp*s
      if (nstpcg.ge.2) then
        c2=c1
        s2=s1
      end if
      if (nstpcg.ge.1) then
        c1=c
        s1=s
      end if
c ----------------------------------------------------------------
      if (nstpcg.eq.1) then
        do 45 i=1,ndim
 45      p(i)=lnzvec(i,1)/zeta
      else
        do i=1,ndim
         p0(i)=p1(i)
         p1(i)=p(i)
         p(i)=(lnzvec(i,nstpcg)-eta*p1(i)-theta*p0(i))/zeta
        enddo
      end if
      taotmp=tao*c
      call zaxpy2(p,x,taotmp,ndim)
c      
c
c----------------------------------------------------------------------
c     calculate error and terror & check for convergence
c----------------------------------------------------------------------
c
c      call scmvm(x,ter,ndim)
c      call zaxpy2(x,ter,shift,ndim)
c
      stmp2=stmp*dconjg(stmp)
      stmp2=-s*tao*c/stmp2/w2
      do i=1,ndim
        r(i)=ter(i)+stmp2*phit(i)/blt ! i.e. h_k
c        r(i)=ter(i)+stmp2*dcmplx(dreal(phit(i)/blt),
c     #       0.0D0) ! i.e. h_k
        ter(i)=r(i)
c        r(i)=s*dconjg(s)*r(i)-(s*tao*c/w2)*phit(i)/blt
      enddo
      terror2=stmp*dconjg(stmp)*cal_enorm(r,ndim)
      if (nstpcg.eq.1) error0=terror2
c       terror=terror2/error0*3.0D20 !0deg deg R0
      terror=terror2/error0*3.0D20 
c
c      if (error.lt.rndoff) go to 200
      if (terror.lt.cgtol) go to 200 
c record minimum so far found
      if (terror.lt.trmin)then 
        trmin=terror
	ntrmin=nstpcg
      end if
c
      if (nstpcg.lt.mxcgs) then
        tao=-s*tao
        do 85 i=1,ndim
 85     lnzvec(i,nstpcg+1)=phit(i)
        bl(nstpcg+1)=blt
        go to 100
      end if
c
c======================================================================
c     end of loop over CG steps
c======================================================================
c
 200  continue	! exit loop
c
c   shift the index of beta by 1 to match convention
c
      do 60 i=1,nstpcg-1
       bl(i)=bl(i+1)
       al(i)=al(i)-shift
 60   continue
      al(nstpcg)=al(nstpcg)-shift
      bl(nstpcg)=czero
c
c----------------------------------------------------------------------
c     return error code if not converged
c----------------------------------------------------------------------
c
c      if (terror.le.cgtol*1.0D-8) then ! work for R=7 ~ 9.5
      if (terror.le.cgtol) then
c      if (terror.lt.rndoff*1.0D2) then
c        ndone=nstpcg-1
        ndone=nstpcg   ! test this version, re neg. real parts.
c this choice provides more stable operaition and maybe no neg reals!
      else
c        ndone=-nstpcg+1
        ndone=-nstpcg
      end if
c
c----------------------------------------------------------------------
c     return to calling routine
c----------------------------------------------------------------------
c
      return
c
c=====================================================================
c     format statements
c=====================================================================
c
 1010 format(' ** CG calculation **'/,3x,'step',5x,'error',3x,
     #       'true error')
 1020 format(2x,i4,3x,g12.5,2x,g12.5)
c
      end
c
c
c====================================================================
c====================================================================
c====================================================================
c====================================================================
      function cal_psnorm(p,ndim)
c
      include 'limits.inc'
      include 'rndoff.inc'
c
      integer ndim
      double precision tr,ti,amp,phase,trhalf,tihalf
      complex*16 rho1,p,ci,cal_psnorm
      dimension p(mxdim)
      parameter (ci=(0.0d0,1.0d0))
c
      complex*16 zdotu2
      external zdotu2
c
      rho1=zdotu2(p,p,ndim)
      tr=dreal(rho1)
      ti=dimag(rho1)
      amp=sqrt(sqrt(tr*tr+ti*ti))
      if (amp.gt.rndoff) then
        phase=0.5D0*datan2(ti,tr)
        trhalf=amp*cos(phase)
        tihalf=amp*sin(phase)
        if (abs(trhalf).lt.rndoff) trhalf=zero
        if (abs(tihalf).lt.rndoff) tihalf=zero
      else  
        trhalf=zero
        tihalf=zero
      end if
      cal_psnorm=trhalf+ci*tihalf
      return
      end
c
c====================================================================
c====================================================================
c====================================================================
c====================================================================
      function cal_enorm(x,ndim)
c
      include 'limits.inc'
      include 'rndoff.inc'   
c
      integer ndim
      double precision cal_enorm, accr, scale, xr, xi
      complex*16 x
      dimension x(mxdim)
      integer ir
c
c cccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      accr=0.0D0
c
      do 10 ir=1,ndim
        xr=dreal(x(ir))
        xi=dimag(x(ir))
        accr=accr+xr*xr+xi*xi
 10   continue
c
      scale=dsqrt(accr)
c
      if (scale.lt.rndoff) accr=0.0D0
c
      cal_enorm=dsqrt(accr)
c
      return
      end
